@@ -127,9 +127,10 @@ poll_for_devices (Scanner *scanner)
     SANE_Status status;
     GList *devices = NULL;
 
+    g_debug ("sane_get_devices ()");
     status = sane_get_devices (&device_list, SANE_FALSE);
     if (status != SANE_STATUS_GOOD) {
-        g_warning ("Unable to get SANE devices: Error %s", sane_strstatus(status));
+        g_warning ("Unable to get SANE devices: %s", sane_strstatus(status));
         return;
     }
 
@@ -153,6 +154,170 @@ poll_for_devices (Scanner *scanner)
 }
 
 
+static void
+set_bool_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Bool value)
+{
+    SANE_Bool v = value;
+    g_return_if_fail (option->type == SANE_TYPE_BOOL);
+    g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %s)", option_index, value ? "TRUE" : "FALSE");
+    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v, NULL);
+}
+
+
+static void
+set_int_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Int value)
+{
+    SANE_Int v = value;
+
+    g_return_if_fail (option->type == SANE_TYPE_INT);
+
+    if (option->constraint_type == SANE_CONSTRAINT_RANGE) {
+        v *= option->constraint.range->quant;
+        if (v < option->constraint.range->min)
+            v = option->constraint.range->min;
+        if (v > option->constraint.range->max)
+            v = option->constraint.range->max;
+    }
+    g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %d)", option_index, value);
+    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v, NULL);
+}
+
+
+static void
+set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, SANE_Fixed value)
+{
+    SANE_Fixed v = value;
+
+    g_return_if_fail (option->type == SANE_TYPE_FIXED);
+
+    if (option->constraint_type == SANE_CONSTRAINT_RANGE) {
+        v *= option->constraint.range->quant;
+        if (v < option->constraint.range->min)
+            v = option->constraint.range->min;
+        if (v > option->constraint.range->max)
+            v = option->constraint.range->max;
+    }
+    g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, %d)", option_index, value);
+    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &v, NULL);
+}
+
+
+static void
+set_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, const char *value)
+{
+    char *string;
+    gsize value_size, size;
+
+    g_return_if_fail (option->type == SANE_TYPE_STRING);
+    
+    value_size = strlen (value) + 1;
+    size = option->size > value_size ? option->size : value_size;
+    string = g_malloc(sizeof(char) * size);
+    strcpy (string, value);
+    g_debug ("sane_control_option (%d, SANE_ACTION_SET_VALUE, \"%s\")", option_index, value);
+    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, string, NULL);
+    g_free (string);
+}
+
+
+static void
+log_option (SANE_Int index, const SANE_Option_Descriptor *option)
+{
+    GString *string;
+    SANE_String_Const *string_iter;
+    SANE_Word i;
+    
+    string = g_string_new ("");
+
+    g_string_append_printf (string, "Option %d: name='%s', title='%s'",
+                            index, option->name, option->title);
+
+    switch (option->type) {
+    case SANE_TYPE_BOOL:
+        g_string_append (string, " type=bool");
+        break;
+    case SANE_TYPE_INT:
+        g_string_append (string, " type=int");
+        break;
+    case SANE_TYPE_FIXED:
+        g_string_append (string, " type=fixed");        
+        break;
+    case SANE_TYPE_STRING:
+        g_string_append (string, " type=string");        
+        break;
+    case SANE_TYPE_BUTTON:
+        g_string_append (string, " type=button");        
+        break;
+    case SANE_TYPE_GROUP:
+        g_string_append (string, " type=group");
+        break;
+    default:
+        g_string_append_printf (string, " type=%d", option->type);
+        break;
+    }
+
+    switch (option->unit) {
+    case SANE_UNIT_NONE:
+        break;
+    case SANE_UNIT_PIXEL:
+        g_string_append (string, " unit=pixels");
+        break;
+    case SANE_UNIT_BIT:
+        g_string_append (string, " unit=bits");
+        break;
+    case SANE_UNIT_MM:
+        g_string_append (string, " unit=mm");
+        break;
+    case SANE_UNIT_DPI:
+        g_string_append (string, " unit=dpi");
+        break;
+    case SANE_UNIT_PERCENT:
+        g_string_append (string, " unit=percent");
+        break;
+    case SANE_UNIT_MICROSECOND:
+        g_string_append (string, " unit=microseconds");
+        break;
+    default:
+        g_string_append_printf (string, " unit=%d", option->unit);
+        break;
+    }
+
+    switch (option->constraint_type) {
+    case SANE_CONSTRAINT_RANGE:
+        g_string_append_printf (string, " min=%d, max=%d, quant=%d",
+                                option->constraint.range->min, option->constraint.range->max,
+                                option->constraint.range->quant);
+        break;
+    case SANE_CONSTRAINT_WORD_LIST:
+        g_string_append (string, " values=[");
+        for (i = 0; i < option->constraint.word_list[0]; i++) {
+            if (i != 0)
+                g_string_append (string, ", ");
+            g_string_append_printf (string, "%d", option->constraint.word_list[i+1]);
+        }
+        g_string_append (string, "]");
+        break;
+    case SANE_CONSTRAINT_STRING_LIST:
+        g_string_append (string, " values=[");
+        for (i = 0; option->constraint.string_list[i]; i++) {
+            if (i != 0)
+                g_string_append (string, ", ");
+            g_string_append_printf (string, "\"%s\"", option->constraint.string_list[i]);
+        }
+        g_string_append (string, "]");
+        break;
+    default:
+        break;
+    }
+
+    g_debug ("%s", string->str);
+    g_string_free (string, TRUE);
+
+    if (option->desc)
+        g_debug ("  Description: %s", option->desc);
+}
+
+
 static gpointer
 scan_thread (Scanner *scanner)
 {
@@ -163,13 +328,15 @@ scan_thread (Scanner *scanner)
     const SANE_Option_Descriptor *option;
     SANE_Int option_index = 0;
     ScanState state = STATE_IDLE;
-    SANE_Int bytes_remaining, line_count = 0, n_read;
+    SANE_Int bytes_remaining, line_count = 0, n_read, page_num;
     SANE_Byte *data;
     SANE_Int version_code;
+    gboolean done;
 
+    g_debug ("sane_init ()");
     status = sane_init (&version_code, NULL);
     if (status != SANE_STATUS_GOOD) {
-        g_warning ("Unable to initialize SANE backend: Error %s", sane_strstatus(status));
+        g_warning ("Unable to initialize SANE backend: %s", sane_strstatus(status));
         return FALSE;
     }
     g_debug ("SANE version %d.%d.%d",
@@ -196,6 +363,7 @@ scan_thread (Scanner *scanner)
             if (device == NULL) {
                 poll_for_devices (scanner);
             } else if (device[0] != '\0') {
+                g_debug ("sane_open ()");
                 status = sane_open (device, &handle);
                 if (status != SANE_STATUS_GOOD) {
                     g_warning ("Unable to get open device: %s", sane_strstatus (status));
@@ -213,52 +381,42 @@ scan_thread (Scanner *scanner)
             break;
 
         case STATE_GET_OPTION:
+            g_debug ("sane_get_option_descriptor (%d)", option_index);
             option = sane_get_option_descriptor (handle, option_index);
             if (!option) {
                 state = STATE_START;
             } else {
-                g_debug ("Option %d: name='%s', title='%s'", option_index, option->name, option->title);
-                if (option->desc)
-                    g_debug ("  Description: %s", option->desc);
-                switch (option->constraint_type) {
-                case SANE_CONSTRAINT_RANGE:
-                    g_debug ("  Range: min=%d, max=%d, quant=%d", option->constraint.range->min, option->constraint.range->max, option->constraint.range->quant);
-                    break;
-                default:
-                    break;
-                }
-                if (option->name && strcmp (option->name, SANE_NAME_SCAN_RESOLUTION) == 0) {
-                    SANE_Int dpi = scanner->priv->dpi;
-                    if (option->constraint_type == SANE_CONSTRAINT_RANGE) {
-                        if (dpi < option->constraint.range->min)
-                            dpi = option->constraint.range->min;
-                        if (dpi > option->constraint.range->max)
-                            dpi = option->constraint.range->max;
+                log_option (option_index, option);
+                if (option->name) {
+                    if (strcmp (option->name, SANE_NAME_SCAN_RESOLUTION) == 0) {
+                        set_fixed_option (handle, option, option_index, scanner->priv->dpi);
                     }
-                    sane_control_option (handle, option_index, SANE_ACTION_SET_VALUE, &dpi, NULL);
                 }
                 option_index++;
             }
             break;
             
         case STATE_START:
+            g_debug ("sane_start ()");
             status = sane_start (handle);
             if (status != SANE_STATUS_GOOD) {
-                g_warning ("Unable to start device: Error %s", sane_strstatus (status));
+                g_warning ("Unable to start device: %s", sane_strstatus (status));
                 emit_signal (scanner, SCAN_FAILED,
                              g_error_new (SCANNER_TYPE, status,
                                           /* Error display when unable to start scan */
                                           _("Unable to start scan")));
                 state = STATE_CLOSE;
             } else {
+                page_num = 0;
                 state = STATE_GET_PARAMETERS;
             }
             break;
             
         case STATE_GET_PARAMETERS:
+            g_debug ("sane_get_parameters ()");
             status = sane_get_parameters (handle, &parameters);
             if (status != SANE_STATUS_GOOD) {
-                g_warning ("Unable to get device parameters: Error %s", sane_strstatus (status));
+                g_warning ("Unable to get device parameters: %s", sane_strstatus (status));
                 emit_signal (scanner, SCAN_FAILED,
                              g_error_new (SCANNER_TYPE, status,
                                           /* Error displayed when communication with scanner broken */
@@ -266,33 +424,45 @@ scan_thread (Scanner *scanner)
                 state = STATE_CLOSE;
             } else {
                 ScanPageInfo *info;
-                
+
                 info = g_malloc(sizeof(ScanPageInfo));
                 info->width = parameters.pixels_per_line;
                 info->height = parameters.lines;
                 info->depth = parameters.depth;
-                emit_signal (scanner, GOT_PAGE_INFO, info);
+
+                if (page_num == 0)
+                    emit_signal (scanner, GOT_PAGE_INFO, info);
+
+                /* Prepare for read */
                 bytes_remaining = parameters.bytes_per_line;
                 data = g_malloc(sizeof(SANE_Byte) * bytes_remaining);
                 line_count = 0;
                 state = STATE_READ;
             }
             break;
-            
+
         case STATE_READ:
+            g_debug ("sane_read (%d)", bytes_remaining);
             status = sane_read (handle, data, bytes_remaining, &n_read);
+            done = FALSE;
+
+            /* End of variable length frame */
             if (status == SANE_STATUS_EOF &&
                 parameters.lines == -1 &&
-                bytes_remaining == parameters.bytes_per_line) {
-                state = STATE_CLOSE;
-            } else if (status != SANE_STATUS_GOOD) {
-                g_warning ("Unable to read frame from device: Error %s", sane_strstatus (status));
+                bytes_remaining == parameters.bytes_per_line)
+                done = TRUE;
+            
+            /* Communication error */
+            else if (status != SANE_STATUS_GOOD) {
+                g_warning ("Unable to read frame from device: %s", sane_strstatus (status));
                 emit_signal (scanner, SCAN_FAILED,
                              g_error_new (SCANNER_TYPE, status,
                                           /* Error displayed when communication with scanner broken */
                                           _("Error communicating with scanner")));
                 state = STATE_CLOSE;
-            } else {
+            }
+            /* Successful read */
+            else {
                 bytes_remaining -= n_read;
                 if (bytes_remaining == 0) {
                     ScanLine *line;
@@ -321,23 +491,37 @@ scan_thread (Scanner *scanner)
                     line->data_length = parameters.bytes_per_line;
                     line->number = line_count;
                     emit_signal (scanner, GOT_LINE, line);
+                    data = NULL;
 
+                    /* On last line */
                     line_count++;
                     if (parameters.lines > 0 && line_count == parameters.lines) {
-                        data = NULL;
-                        state = STATE_CLOSE;
-                    } else {
+                        done = TRUE;
+                    }
+                    else {
                         bytes_remaining = parameters.bytes_per_line;
                         data = g_malloc(sizeof(SANE_Byte) * bytes_remaining);
                     }
                 }
             }
+
+            /* End scan or start next frame */
+            if (done) {
+                if (parameters.last_frame)
+                    state = STATE_CLOSE;
+                else {
+                    page_num++;
+                    state = STATE_GET_PARAMETERS;
+                }
+            }
             break;
             
         case STATE_CLOSE:
             emit_signal (scanner, IMAGE_DONE, NULL);
-            if (handle)
+            if (handle) {
+                g_debug ("sane_close ()");
                 sane_close (handle);
+            }
             handle = NULL;
             g_free (data);
             data = NULL;
@@ -360,6 +544,18 @@ scanner_new ()
 }
 
 
+void
+scanner_start (Scanner *scanner)
+{
+    GError *error = NULL;
+    scanner->priv->thread = g_thread_create ((GThreadFunc) scan_thread, scanner, TRUE, &error);
+    if (error) {
+        g_critical ("Unable to create thread: %s", error->message);
+        g_error_free (error);
+    }    
+}
+
+
 void
 scanner_scan (Scanner *scanner, const char *device, gint dpi_)
 {
@@ -380,14 +576,14 @@ void scanner_free (Scanner *scanner)
     g_debug ("Stopping scan thread");
     scanner->priv->running = FALSE;
     g_async_queue_push (scanner->priv->scan_queue, "");
-    g_thread_join (scanner->priv->thread);
+    if (scanner->priv->thread)
+        g_thread_join (scanner->priv->thread);
 
     g_async_queue_unref (scanner->priv->scan_queue);
     g_object_unref (scanner);
 
-    g_debug ("Closing SANE interface");
+    g_debug ("sane_exit ()");
     sane_exit ();
-    g_debug ("SANE interface closed");
 }
 
 
@@ -450,15 +646,7 @@ scanner_class_init (ScannerClass *klass)
 static void
 scanner_init (Scanner *scanner)
 {
-    GError *error = NULL;
-    
     scanner->priv = G_TYPE_INSTANCE_GET_PRIVATE (scanner, SCANNER_TYPE, ScannerPrivate);
-
     scanner->priv->running = TRUE;
     scanner->priv->scan_queue = g_async_queue_new ();
-    scanner->priv->thread = g_thread_create ((GThreadFunc) scan_thread, scanner, TRUE, &error);
-    if (error) {
-        g_critical ("Unable to create thread: %s", error->message);
-        g_error_free (error);
-    }
 }
