@@ -1,4 +1,17 @@
+/*
+ * Copyright (C) 2009 Canonical Ltd.
+ * Author: Robert Ancell <robert.ancell@canonical.com>
+ * 
+ * This program is free software: you can redistribute it and/or modify it under
+ * the terms of the GNU General Public License as published by the Free Software
+ * Foundation, either version 3 of the License, or (at your option) any later
+ * version. See http://www.gnu.org/copyleft/gpl.html the full text of the
+ * license.
+ */
+
+#include <stdlib.h>
 #include <string.h>
+#include <glib/gi18n.h>
 #include <gtk/gtk.h>
 #include <gconf/gconf-client.h>
 
@@ -11,6 +24,7 @@ enum {
     STOP_SCAN,
     SAVE,
     PRINT,
+    QUIT,
     LAST_SIGNAL
 };
 static guint signals[LAST_SIGNAL] = { 0, };
@@ -24,8 +38,10 @@ struct SimpleScanPrivate
     GtkWidget *scan_label;
     GtkWidget *actions_box;
     GtkWidget *device_combo, *mode_combo;
+    GtkTreeModel *device_model, *mode_model;
     GtkWidget *preview_area;
 
+    gchar *default_file_name;
     gboolean scanning;
     Orientation orientation;
 };
@@ -36,18 +52,16 @@ G_DEFINE_TYPE (SimpleScan, ui, G_TYPE_OBJECT);
 static gboolean
 find_scan_device (SimpleScan *ui, const char *device, GtkTreeIter *iter)
 {
-    GtkTreeModel *model;
     gboolean have_iter = FALSE;
 
-    model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->device_combo));
-    if (gtk_tree_model_get_iter_first (model, iter)) {
+    if (gtk_tree_model_get_iter_first (ui->priv->device_model, iter)) {
         do {
             gchar *d;
-            gtk_tree_model_get (model, iter, 0, &d, -1);
+            gtk_tree_model_get (ui->priv->device_model, iter, 0, &d, -1);
             if (strcmp (d, device) == 0)
                 have_iter = TRUE;
             g_free (d);
-        } while (!have_iter && gtk_tree_model_iter_next (model, iter));
+        } while (!have_iter && gtk_tree_model_iter_next (ui->priv->device_model, iter));
     }
     
     return have_iter;
@@ -60,11 +74,8 @@ get_selected_device (SimpleScan *ui)
     GtkTreeIter iter;
 
     if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (ui->priv->device_combo), &iter)) {
-        GtkTreeModel *model;
         gchar *device;
-
-        model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->device_combo));
-        gtk_tree_model_get (model, &iter, 0, &device, -1);
+        gtk_tree_model_get (ui->priv->device_model, &iter, 0, &device, -1);
         return device;
     }
 
@@ -72,17 +83,23 @@ get_selected_device (SimpleScan *ui)
 }
 
 
+void
+ui_set_default_file_name (SimpleScan *ui, const gchar *default_file_name)
+{
+    g_free (ui->priv->default_file_name);
+    ui->priv->default_file_name = g_strdup (default_file_name);
+}
+
+
 void
 ui_mark_devices_undetected (SimpleScan *ui)
 {
-    GtkTreeModel *model;
     GtkTreeIter iter;
     
-    model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->device_combo));
-    if (gtk_tree_model_get_iter_first (model, &iter)) {
+    if (gtk_tree_model_get_iter_first (ui->priv->device_model, &iter)) {
         do {
-            gtk_list_store_set (GTK_LIST_STORE (model), &iter, 2, FALSE, -1);
-        } while (gtk_tree_model_iter_next (model, &iter));
+            gtk_list_store_set (GTK_LIST_STORE (ui->priv->device_model), &iter, 2, FALSE, -1);
+        } while (gtk_tree_model_iter_next (ui->priv->device_model, &iter));
     }
 }
 
@@ -91,16 +108,13 @@ void
 ui_add_scan_device (SimpleScan *ui, const gchar *device, const gchar *label)
 {
     GtkTreeIter iter;
-    GtkTreeModel *model;
-
-    model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->device_combo));
     
     if (!find_scan_device (ui, device, &iter)) {
-        gtk_list_store_append (GTK_LIST_STORE (model), &iter);
-        gtk_list_store_set (GTK_LIST_STORE (model), &iter, 0, device, -1);
+        gtk_list_store_append (GTK_LIST_STORE (ui->priv->device_model), &iter);
+        gtk_list_store_set (GTK_LIST_STORE (ui->priv->device_model), &iter, 0, device, -1);
     }
 
-    gtk_list_store_set (GTK_LIST_STORE (model), &iter, 1, label, 2, TRUE, -1);
+    gtk_list_store_set (GTK_LIST_STORE (ui->priv->device_model), &iter, 1, label, 2, TRUE, -1);
     
     /* Select this device if none selected */
     if (gtk_combo_box_get_active (GTK_COMBO_BOX (ui->priv->device_combo)) == -1)
@@ -108,19 +122,54 @@ ui_add_scan_device (SimpleScan *ui, const gchar *device, const gchar *label)
 }
 
 
-static void
-get_document_hint (SimpleScan *ui)
+void
+ui_set_selected_device (SimpleScan *ui, const gchar *device)
 {
     GtkTreeIter iter;
 
-    if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (ui->priv->mode_combo), &iter)) {
-        GtkTreeModel *model;
-        gchar *mode;
-
-        model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->mode_combo));
-        gtk_tree_model_get (model, &iter, 0, &mode, -1);
-        g_free (mode);
+    /* If doesn't exist add with label set to device name */
+    if (!find_scan_device (ui, device, &iter)) {
+        gtk_list_store_append (GTK_LIST_STORE (ui->priv->device_model), &iter);
+        gtk_list_store_set (GTK_LIST_STORE (ui->priv->device_model), &iter, 0, device, 1, device, 2, FALSE, -1);
     }
+
+    gtk_combo_box_set_active_iter (GTK_COMBO_BOX (ui->priv->device_combo), &iter);
+}
+
+
+static void
+set_document_hint (SimpleScan *ui, const gchar *document_hint)
+{
+    GtkTreeIter iter;
+
+    if (gtk_tree_model_get_iter_first (ui->priv->mode_model, &iter)) {
+        do {
+            gchar *d;
+            gboolean have_match;
+
+            gtk_tree_model_get (ui->priv->mode_model, &iter, 0, &d, -1);
+            have_match = strcmp (d, document_hint) == 0;
+            g_free (d);
+
+            if (have_match) {
+                gtk_combo_box_set_active_iter (GTK_COMBO_BOX (ui->priv->mode_combo), &iter);                
+                return;
+            }
+        } while (gtk_tree_model_iter_next (ui->priv->mode_model, &iter));
+     }
+}
+
+
+static gchar *
+get_document_hint (SimpleScan *ui)
+{
+    GtkTreeIter iter;
+    gchar *mode = NULL;
+
+    if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (ui->priv->mode_combo), &iter))
+        gtk_tree_model_get (ui->priv->mode_model, &iter, 0, &mode, -1);
+    
+    return mode;
 }
 
 
@@ -150,11 +199,14 @@ scan_button_clicked_cb (GtkWidget *widget, SimpleScan *ui)
     if (ui->priv->scanning) {
         g_signal_emit (G_OBJECT (ui), signals[STOP_SCAN], 0);
     } else {
-        gchar *device;
+        gchar *device, *mode;
+
         device = get_selected_device (ui);
         if (device) {
-            g_signal_emit (G_OBJECT (ui), signals[START_SCAN], 0, device);
+            mode = get_document_hint (ui);
+            g_signal_emit (G_OBJECT (ui), signals[START_SCAN], 0, device, mode);
             g_free (device);
+            g_free (mode);
         }
     }
 }
@@ -178,14 +230,16 @@ save_file_button_clicked_cb (GtkWidget *widget, SimpleScan *ui)
     GtkWidget *dialog;
     gint response;
 
-    dialog = gtk_file_chooser_dialog_new ("Save As...", GTK_WINDOW (ui->priv->window),
+    /* Title of save dialog */
+    dialog = gtk_file_chooser_dialog_new (_("Save As..."),
+                                          GTK_WINDOW (ui->priv->window),
                                           GTK_FILE_CHOOSER_ACTION_SAVE,
                                           GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                           GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
                                           NULL);
     gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
     gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (dialog), FALSE);
-    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), "Scanned Document.pdf");
+    gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), ui->priv->default_file_name);
     
     response = gtk_dialog_run (GTK_DIALOG (dialog));
     if (response == GTK_RESPONSE_ACCEPT) {
@@ -251,7 +305,8 @@ load_device_cache (SimpleScan *ui)
     key_file = g_key_file_new ();
     result = g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, &error);
     if (error) {
-        g_warning ("Error loading device cache file: %s", error->message);
+        if (!g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
+            g_warning ("Error loading device cache file: %s", error->message);
         g_error_free (error);
         error = NULL;
     }
@@ -292,7 +347,7 @@ save_device_cache (SimpleScan *ui)
 
     g_debug ("Saving device cache");
 
-    model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->device_combo));
+    model = ui->priv->device_model;
     if (gtk_tree_model_get_iter_first (model, &iter)) {
         GKeyFile *key_file;
         gchar *data;
@@ -357,32 +412,43 @@ G_MODULE_EXPORT
 gboolean
 window_delete_event_cb (GtkWidget *widget, GdkEvent *event, SimpleScan *ui)
 {
-    char *device;
+    char *device, *document_type;
+
     save_device_cache (ui);
+
     device = get_selected_device (ui);
     if (device) {
         gconf_client_set_string(ui->priv->client, "/apps/simple-scan/selected_device", device, NULL);
         g_free (device);
     }
-    gtk_main_quit ();
+
+    document_type = get_document_hint (ui);
+    gconf_client_set_string(ui->priv->client, "/apps/simple-scan/document_type", document_type, NULL);
+    g_free (document_type);
+
+    g_signal_emit (G_OBJECT (ui), signals[QUIT], 0);
     return TRUE;
 }
 
 
-static gboolean
+static void
 ui_load (SimpleScan *ui)
 {
     GtkBuilder *builder;
     GError *error = NULL;
     GtkCellRenderer *renderer;
-    GtkTreeIter iter;
-    gchar *device;
+    gchar *device, *document_type;
 
     builder = gtk_builder_new ();
     gtk_builder_add_from_file (builder, UI_DIR "simple-scan.ui", &error);
     if (error) {
         g_critical ("Unable to load UI: %s\n", error->message);
-        return FALSE;
+        ui_show_error (ui,
+                       /* Title of dialog when cannot load required files */
+                       _("Files missing"),
+                       /* Description in dialog when cannot load required files */
+                       _("Please check your installation"));
+        exit (1);
     }
     gtk_builder_connect_signals (builder, ui);
 
@@ -390,7 +456,9 @@ ui_load (SimpleScan *ui)
     ui->priv->scan_label = GTK_WIDGET (gtk_builder_get_object (builder, "scan_label"));
     ui->priv->actions_box = GTK_WIDGET (gtk_builder_get_object (builder, "actions_box"));
     ui->priv->device_combo = GTK_WIDGET (gtk_builder_get_object (builder, "device_combo"));
+    ui->priv->device_model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->device_combo));
     ui->priv->mode_combo = GTK_WIDGET (gtk_builder_get_object (builder, "mode_combo"));
+    ui->priv->mode_model = gtk_combo_box_get_model (GTK_COMBO_BOX (ui->priv->mode_combo));
     ui->priv->preview_area = GTK_WIDGET (gtk_builder_get_object (builder, "preview_area"));
 
     renderer = gtk_cell_renderer_text_new();
@@ -405,11 +473,18 @@ ui_load (SimpleScan *ui)
     /* Load previously detected scanners and select the last used one */
     load_device_cache (ui);
     device = gconf_client_get_string(ui->priv->client, "/apps/simple-scan/selected_device", NULL);
-    if (device && find_scan_device (ui, device, &iter)) {
-        gtk_combo_box_set_active_iter (GTK_COMBO_BOX (ui->priv->device_combo), &iter);
+    if (device) {
+        GtkTreeIter iter;
+        if (find_scan_device (ui, device, &iter))
+            gtk_combo_box_set_active_iter (GTK_COMBO_BOX (ui->priv->device_combo), &iter);
+        g_free (device);
+    }
+    
+    document_type = gconf_client_get_string(ui->priv->client, "/apps/simple-scan/document_type", NULL);
+    if (document_type) {
+        set_document_hint (ui, document_type);
+        g_free (document_type);
     }
-
-    return TRUE;
 }
 
 
@@ -425,9 +500,13 @@ ui_set_scanning (SimpleScan *ui, gboolean scanning)
 {
     ui->priv->scanning = scanning;
     if (ui->priv->scanning)
-        gtk_label_set_label (GTK_LABEL (ui->priv->scan_label), "_Cancel");
+        gtk_label_set_label (GTK_LABEL (ui->priv->scan_label),
+                             /* Label on cancel scan button */
+                             _("_Cancel"));
     else
-        gtk_label_set_label (GTK_LABEL (ui->priv->scan_label), "_Scan");
+        gtk_label_set_label (GTK_LABEL (ui->priv->scan_label),
+                             /* Label on scan button */
+                             _("_Scan"));
 }
 
 
@@ -451,6 +530,37 @@ ui_redraw_preview (SimpleScan *ui)
 }
 
 
+void
+ui_show_error (SimpleScan *ui, const gchar *error_title, const gchar *error_text)
+{
+    GtkWidget *dialog;
+
+    dialog = gtk_message_dialog_new (GTK_WINDOW (ui->priv->window),
+                                     GTK_DIALOG_MODAL,
+                                     GTK_MESSAGE_WARNING,
+                                     GTK_BUTTONS_CLOSE,
+                                     "%s", error_title);
+    gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
+                                              "%s", error_text);
+
+    gtk_dialog_run (GTK_DIALOG (dialog));
+    gtk_widget_destroy (dialog);
+}
+
+
+void
+ui_start (SimpleScan *ui)
+{
+    if (gtk_tree_model_iter_n_children (ui->priv->device_model, NULL) == 0)
+        ui_show_error (ui,
+                       /* Warning displayed when no scanners are detected */
+                       _("No scanners detected"),
+                       /* Hint to user on why there are no scanners detected */
+                       _("Please check your scanner is connected and powered on"));
+}
+
+
+/* Generated with glib-genmarshal */
 static void
 g_cclosure_user_marshal_VOID__POINTER_DOUBLE_DOUBLE (GClosure     *closure,
                                                      GValue       *return_value G_GNUC_UNUSED,
@@ -490,6 +600,44 @@ g_cclosure_user_marshal_VOID__POINTER_DOUBLE_DOUBLE (GClosure     *closure,
 }
 
 
+/* Generated with glib-genmarshal */
+void
+g_cclosure_user_marshal_VOID__STRING_STRING (GClosure     *closure,
+                                             GValue       *return_value G_GNUC_UNUSED,
+                                             guint         n_param_values,
+                                             const GValue *param_values,
+                                             gpointer      invocation_hint G_GNUC_UNUSED,
+                                             gpointer      marshal_data)
+{
+  typedef void (*GMarshalFunc_VOID__STRING_STRING) (gpointer       data1,
+                                                    gconstpointer  arg_1,
+                                                    gconstpointer  arg_2,
+                                                    gpointer       data2);
+  register GMarshalFunc_VOID__STRING_STRING callback;
+  register GCClosure *cc = (GCClosure*) closure;
+  register gpointer data1, data2;
+
+  g_return_if_fail (n_param_values == 3);
+
+  if (G_CCLOSURE_SWAP_DATA (closure))
+    {
+      data1 = closure->data;
+      data2 = g_value_peek_pointer (param_values + 0);
+    }
+  else
+    {
+      data1 = g_value_peek_pointer (param_values + 0);
+      data2 = closure->data;
+    }
+  callback = (GMarshalFunc_VOID__STRING_STRING) (marshal_data ? marshal_data : cc->callback);
+
+  callback (data1,
+            g_value_get_string (param_values + 1),
+            g_value_get_string (param_values + 2),
+            data2);
+}
+
+
 static void
 ui_class_init (SimpleScanClass *klass)
 {
@@ -507,8 +655,8 @@ ui_class_init (SimpleScanClass *klass)
                       G_SIGNAL_RUN_LAST,
                       G_STRUCT_OFFSET (SimpleScanClass, start_scan),
                       NULL, NULL,
-                      g_cclosure_marshal_VOID__INT,
-                      G_TYPE_NONE, 1, G_TYPE_INT);
+                      g_cclosure_user_marshal_VOID__STRING_STRING,
+                      G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);
     signals[STOP_SCAN] =
         g_signal_new ("stop-scan",
                       G_TYPE_FROM_CLASS (klass),
@@ -533,6 +681,14 @@ ui_class_init (SimpleScanClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__POINTER,
                       G_TYPE_NONE, 1, G_TYPE_POINTER);
+    signals[QUIT] =
+        g_signal_new ("quit",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (SimpleScanClass, quit),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
 
     g_type_class_add_private (klass, sizeof (SimpleScanPrivate));
 }
@@ -545,7 +701,8 @@ ui_init (SimpleScan *ui)
 
     ui->priv->client = gconf_client_get_default();
     gconf_client_add_dir(ui->priv->client, "/apps/simple-scan", GCONF_CLIENT_PRELOAD_NONE, NULL);
-    
+
+    ui->priv->default_file_name = g_strdup (_("Scanned Document.pdf"));
     ui->priv->scanning = FALSE;
     ui->priv->orientation = TOP_TO_BOTTOM;
 
